#include "stdio.h"
#include "string.h"
#include "math.h"
#include "stdlib.h"

#include "SDL.h"
#include "SDL_net.h"
#include "SDL_rotozoom.h"
#include "SDL_ttf.h"
#include "SDL_mixer.h"
#include "SDL_joystick.h"

#include "BList.h"

#include "samples.h"
#include "animscript.h"
#include "conditionscript.h"
#include "actionscript.h"
#include "auxiliar.h"
#include "MOG2_auxiliar.h"
#include "variables.h"
#include "mog2.h"
#include "tiles.h"
#include "object.h"
#include "interfaces.h"
#include "mog2multiplayer.h"
#include "activeobject.h"
#include "activeroom.h"
#include "activeMAP.h"
#include "activemog2.h"

#include "randomc.h"

#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif

int MAX_HYPOTHESES=8;
const int MAX_CONNECTIONS=256;
const int CONNECTION_TIMEOUT=1000;
const int LISTENING_TIME=1;

const int MSG_MAXLEN=2048;


/* Gomplete Network game configuration: */ 

NWMOG2GameInfo::NWMOG2GameInfo(MOG2 *g) 
{
	game=g;
	port=32123;
	socket_set=0;
	server_socket=0;
	server=0;
	screen_x=640;
	screen_y=480;
	render=0;

	local=new NWMOG2ClientInfo();
	local->LVK_ID=-1;
	strcpy(local->computer_name,"my_computer");
	strcpy(local->computer_address,"");
	local->socket=0;

} /* NWMOG2GameInfo::NWMOG2GameInfo */ 


NWMOG2GameInfo::~NWMOG2GameInfo()
{
	if (socket_set!=0) {
		SDLNet_FreeSocketSet(socket_set);
		socket_set=0;
	} /* if */ 

	if (server!=local) delete server;

	clients.DeleteElement(local);
	delete local;
} /* NWMOG2GameInfo::~NWMOG2GameInfo */ 


NWMOG2ClientInfo::NWMOG2ClientInfo()
{
	socket=0;
	computer_name[0]=0;
	computer_address[0]=0;
} /* NWMOG2ClientInfo::NWMOG2ClientInfo */ 


NWMOG2ClientInfo::~NWMOG2ClientInfo()
{
	if (socket!=0) SDLNet_TCP_Close(socket);
	socket=0;
} /* NWMOG2ClientInfo::~NWMOG2ClientInfo */ 


int NWMOG2GameInfo::new_LVK_identifier(void)
{	
	int ID=0;

	while(find_by_LVK_identifier(ID)!=0) ID++;

	return ID;
} /* NWMOG2GameInfo::new_LVK_identifier */ 


int NWMOG2GameInfo::new_player_identifier(void)
{
	int ID=0;

	while(find_by_player_identifier(ID)!=0) ID++;

	return ID;
} /* NWMOG2GameInfo::new_player_identifier */ 



NWMOG2ClientInfo *NWMOG2GameInfo::find_by_LVK_identifier(int ID)
{
	List<NWMOG2ClientInfo> l;
	NWMOG2ClientInfo *client;

	l.Instance(clients);
	l.Rewind();
	while(l.Iterate(client)) {
		if (client->LVK_ID==ID) return client;
	} /* while */ 

	if (server->LVK_ID==ID) return server;
	if (local->LVK_ID==ID) return local;

	return 0;
} /* NWMOG2ClientInfo::find_by_LVK_identifier */ 


NWMOG2PlayerInfo *NWMOG2GameInfo::find_by_player_identifier(int ID)
{
	List<NWMOG2ClientInfo> l;
	NWMOG2ClientInfo *client;
	NWMOG2PlayerInfo *player;

	l.Instance(clients);
	l.Rewind();
	while(l.Iterate(client)) {
		player=client->find_by_player_identifier(ID);
		if (player!=0) return player;
	} /* while */ 

	if (server!=0) {
		player=server->find_by_player_identifier(ID);
		if (player!=0) return player;
	} /* if */ 

	if (local!=0) {
		player=local->find_by_player_identifier(ID);
		if (player!=0) return player;
	} /* if */ 

	return 0;
} /* NWMOG2GameInfo::find_by_player_ID */ 


NWMOG2PlayerInfo *NWMOG2ClientInfo::find_by_player_identifier(int ID)
{
	List<NWMOG2PlayerInfo> l;
	NWMOG2PlayerInfo *player;

	l.Instance(players);
	l.Rewind();
	while(l.Iterate(player)) {
		if (player->player_ID==ID) return player;
	} /* while */ 

	return 0;
} /* NWMOG2ClientInfo::find_by_player_ID */ 


int NWMOG2GameInfo::get_local_nplayers(void)
{
	return local->get_nplayers();
} /* NWMOG2GameInfo::get_local_nplayers */ 


int NWMOG2GameInfo::get_nplayers(void)
{
	int n=0;
	List<NWMOG2ClientInfo> l;
	NWMOG2ClientInfo *client;

	l.Instance(clients);
	l.Rewind();
	while(l.Iterate(client)) if (client!=local) n+=client->get_nplayers();

	if (server!=0 && server!=local) n+=server->get_nplayers();
	if (local!=0) n+=local->get_nplayers();

	return n;
} /* NWMOG2GameInfo::get_nplayers */ 


int NWMOG2ClientInfo::get_nplayers(void)
{
	return players.Length();
} /* NWMOG2ClientInfo::get_nplayers */ 



NetworkMOG2Protocol::NetworkMOG2Protocol(NWMOG2GameInfo *gi,int gm,char *file)
{
	int i;

	gameinfo=gi;

	LVK=new KEYBOARDSTATE *[gameinfo->clients.Length()+1];
	LVK_cycle=new int[gameinfo->clients.Length()+1];
	hypothetic_LVK=new List<KEYBOARDSTATE>[gameinfo->clients.Length()+1];
	
	client_active=new bool[gameinfo->clients.Length()+1];

	for(i=0;i<gameinfo->clients.Length()+1;i++) {
		LVK[i]=new KEYBOARDSTATE();
		LVK[i]->LVK_ID=i;
		LVK_cycle[i]=(-1)&0xff;
		client_active[i]=false;
	} /* for */ 
	client_active[gameinfo->local->LVK_ID]=true;

	current_cycle=0;
	last_synchronized_cycle=-1;
	connection_state=0;

	bloqued=false;

	game_mode=gm;
	replay_file=0;
	switch(game_mode) {
	case 0:
		replay_file=0;
		break;
	case 1:
		/* ERRORR!!!! */ 
		break;
	case 2:
		{
			int i;
			NWMOG2PlayerInfo *pi;
			char *tmp;

			replay_file=fopen(file,"w");
			fprintf(replay_file,"%i %i\n",gameinfo->get_nplayers(),gameinfo->clients.Length()+1);

			for(i=0;i<gameinfo->get_nplayers();i++) {
				pi=gameinfo->find_by_player_identifier(i);
				
				encode_string(pi->avatar->name,&tmp);
				fprintf(replay_file,"%s\n",tmp);
				delete tmp;

				encode_string(pi->name,&tmp);
				fprintf(replay_file,"%s\n",tmp);
				delete tmp;

				fprintf(replay_file,"%i %i\n",pi->start_room->get_x(),pi->start_room->get_y());

				encode_string(pi->start_map->name,&tmp);
				fprintf(replay_file,"%s\n",tmp);
				delete tmp;

				fprintf(replay_file,"%i %i %i %i %f\n",pi->vp_area.x,pi->vp_area.y,
													   pi->vp_area.w,pi->vp_area.h,
													   pi->vp_zoom);

				if (pi->start_interface!=0) {
					encode_string(pi->start_interface->name,&tmp);
					fprintf(replay_file,"INTERFACE %s\n",tmp);
					delete tmp;
				} else {
					fprintf(replay_file,"NOINTERFACE\n");
				} /* if */ 

				fprintf(replay_file,"%i %i %i %i %i %i %i %i\n",pi->left_key,pi->right_key,pi->up_key,pi->down_key,
																pi->fire1_key,pi->fire2_key,pi->fire3_key,pi->fire4_key);
			} /* for */ 
		}

		break;
	} /* switch */ 

} /* NetworkMOG2Protocol::NetworkMOG2Protocol */ 


NetworkMOG2Protocol::~NetworkMOG2Protocol()
{
	int i;
	delete client_active;

	for(i=0;i<gameinfo->clients.Length()+1;i++) {
		delete LVK[i];
		LVK[i]=0;
	} /* for */ 
	delete LVK;
	delete LVK_cycle;
	delete []hypothetic_LVK;

	switch(game_mode) {
	case 0:
		break;
	case 1:
		/* ERRORR!!!! */ 
		break;
	case 2:
		fprintf(replay_file,"END\n");
		fclose(replay_file);
		break;
	} /* switch */ 

} /*NetworkMOG2Protocol::~NetworkMOG2Protocol */ 



bool NetworkMOG2Protocol::listening_cycle(void)
{
	bool retval=true;
	char data[MSG_MAXLEN];
	int msg_len;

	/* Listen Network: */ 
	if (gameinfo->server==gameinfo->local) {
		bool first=true;
		bool all_present=true;
		int i,ncomputers;
		for(ncomputers=0,i=0;i<gameinfo->clients.Length()+1;i++) if (client_active[i]) ncomputers++;

		/* SERVER CODE: */ 
		while(retval &&
			(SDLNet_CheckSockets(gameinfo->socket_set, LISTENING_TIME)!=0 ||
			 (ncomputers==1 && first))) {
			List<NWMOG2ClientInfo> l;
			NWMOG2ClientInfo *ci;	

			first=false;

			l.Instance(gameinfo->clients);
			l.Rewind();
			while(l.Iterate(ci)) {
				if (SDLNet_SocketReady(ci->socket)) {
					if (SDLNet_TCP_Recv(ci->socket,data,4)!=4) {
						retval=false;
					} else {
						switch(SDLNet_Read16(data)) {
						case 17:
							// Client Starting:
							if (!client_active[ci->LVK_ID]) {
								client_active[ci->LVK_ID]=true;
								// Send client the current cycle: 
								
								SDLNet_Write16(18,data);
								SDLNet_Write16((last_synchronized_cycle+1)&0xff,data+2);
								SDLNet_TCP_Send(ci->socket,data,4);

							} /* if */ 
							break;

						case 26:
							// Client Disconnectiong:
							client_active[ci->LVK_ID]=false;
							break;

						case 21:
							// Client is bloqued!!!
							/* ... */ 
							break;

						case 20:
							{
								int i,pos;
								int cycle;
								KEYBOARDSTATE *lvk;

								msg_len=SDLNet_Read16(data+2);

								if (SDLNet_TCP_Recv(ci->socket,data,msg_len)!=msg_len) {
									retval=false;
								} else {

									// Receiving a client's LVK: 
									lvk=new KEYBOARDSTATE();
									
									lvk->LVK_ID=SDLNet_Read16(data);
									cycle=SDLNet_Read16(data+2);
									pos=4;

									for(i=0;i<SDLK_LAST;i++) {
										lvk->keyboard[i]=LVK[lvk->LVK_ID]->keyboard[i];
										lvk->old_keyboard[i]=LVK[lvk->LVK_ID]->keyboard[i];
										LVK[lvk->LVK_ID]->old_keyboard[i]=LVK[lvk->LVK_ID]->keyboard[i];
									} /* for */ 
									LVK_cycle[lvk->LVK_ID]=cycle;
									LVK[lvk->LVK_ID]->keyevents.Delete();
									
									while(pos<msg_len) {
										switch(data[pos++]) {
										case 0:i=SDLNet_Read16(data+pos);
											   pos+=2;
											   lvk->keyboard[i]=1;
											   LVK[lvk->LVK_ID]->keyboard[i]=1;
											break;
										case 1:i=SDLNet_Read16(data+pos);
											   pos+=2;
											   lvk->keyboard[i]=0;
											   LVK[lvk->LVK_ID]->keyboard[i]=0;
											break;
										case 2:{
													SDL_keysym *ks;

													ks=new SDL_keysym();
													ks->scancode=(unsigned char)SDLNet_Read16(data+pos);
													ks->sym=(SDLKey)SDLNet_Read16(data+pos+2);
													ks->mod=(SDLMod)SDLNet_Read16(data+pos+4);
													ks->unicode=SDLNet_Read16(data+pos+6);
													lvk->keyevents.Add(ks);

													ks=new SDL_keysym();
													ks->scancode=(unsigned char)SDLNet_Read16(data+pos);
													ks->sym=(SDLKey)SDLNet_Read16(data+pos+2);
													ks->mod=(SDLMod)SDLNet_Read16(data+pos+4);
													ks->unicode=SDLNet_Read16(data+pos+6);
													LVK[lvk->LVK_ID]->keyevents.Add(ks);
													pos+=8;
											   }
											break;
										} /* switch */ 
									} /* while */ 	
									
									hypothetic_LVK[lvk->LVK_ID].Add(lvk);
								} /* if */ 
							} 
							break;
						} /* switch */ 
					} /* if */ 
				} /* if */ 
			} /* while */ 
		} /* while */ 

		// Check for complete cycles: 
		do {
			int i;
			
			all_present=true;
			for(i=0;i<gameinfo->clients.Length()+1;i++) {
				if (hypothetic_LVK[i].EmptyP()) all_present=false;
			} /* for */ 

			if (all_present) {
				// cycle complete!:

				// Send all the LVKs to the clients:
				List<NWMOG2ClientInfo> l;
				NWMOG2ClientInfo *ci;
				int i,j,next_cycle=(last_synchronized_cycle+1)&0xff;
				int pos=0;
				List<SDL_keysym> l2;
				SDL_keysym *ks;
				KEYBOARDSTATE *lvk;

				l.Instance(gameinfo->clients);
				l.Rewind();
				while(l.Iterate(ci)) {
					if (client_active[ci->LVK_ID]) {
						for(j=0;j<gameinfo->clients.Length()+1;j++) {

							if (client_active[j]) {
								lvk=hypothetic_LVK[j][0];
								SDLNet_Write16(20,data);
								SDLNet_Write16(j,data+4);
								SDLNet_Write16(next_cycle,data+6);
								pos=8;

								for(i=0;i<SDLK_LAST;i++) {
									if (lvk->keyboard[i]!=lvk->old_keyboard[i]) {
										if (lvk->keyboard[i] && !lvk->old_keyboard[i]) {
											data[pos++]=0;
											SDLNet_Write16(i,data+pos);
											pos+=2;
										} else {
											data[pos++]=1;
											SDLNet_Write16(i,data+pos);
											pos+=2;
										} /* if */ 
									} /* if */ 
								} /* for */ 

								l2.Instance(lvk->keyevents);
								l2.Rewind();
								while(l2.Iterate(ks)) {
									data[pos++]=2;
									SDLNet_Write16(ks->scancode,data+pos);
									SDLNet_Write16(ks->sym,data+pos+2);
									SDLNet_Write16(ks->mod,data+pos+4);
									SDLNet_Write16(ks->unicode,data+pos+6);
									pos+=8;
								} /* while */ 

								SDLNet_Write16(pos-4,data+2);
								SDLNet_TCP_Send(ci->socket,data,pos);
							} else {
								SDLNet_Write16(20,data);
								SDLNet_Write16(4,data+2);
								SDLNet_Write16(j,data+4);
								SDLNet_Write16(next_cycle,data+6);

								SDLNet_TCP_Send(ci->socket,data,8);
							} /* if */ 
						} /* for */ 	
					} /* if */ 
				} /* while */ 
				
				// Update wrong hypotheses: 
				{
					KEYBOARDSTATE **LVKs;

					LVKs=new KEYBOARDSTATE *[gameinfo->clients.Length()+1];

					for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]=hypothetic_LVK[i][0];
					game_update_hypotheses(LVKs,gameinfo->clients.Length()+1,next_cycle);
					for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]=0;

					if (game_mode==2) {
						fprintf(replay_file,"NEWCYCLE\n",next_cycle);
						for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]->save(replay_file);
					} /* if */ 
					
					delete LVKs;
				}

				for(i=0;i<gameinfo->clients.Length()+1;i++) delete hypothetic_LVK[i].ExtractIni();

				last_synchronized_cycle=next_cycle;
			} /* if */ 
		} while(all_present);

	} else {

		/* CLIENT CODE: */
		bool new_lvk;
		switch(connection_state) {
		case 0:
				// Client still not registered: 
				SDLNet_Write16(17,data);
				SDLNet_Write16(0,data+2);
				SDLNet_TCP_Send(gameinfo->server_socket,data,4);
				
				/* set timeout: */ 
				network_timeout_start=SDL_GetTicks();
				network_timeout=CONNECTION_TIMEOUT;	
				connection_state=1;
				break;

		case 1:
				// Awaiting Server confirmation: 
				if (SDL_GetTicks()-network_timeout_start>network_timeout) {
					/* TIMEOUT: */ 	
					connection_state=0;
				} else {
					if (SDLNet_CheckSockets(gameinfo->socket_set, LISTENING_TIME)!=0) {
						if (SDLNet_SocketReady(gameinfo->server_socket)) {
							/* Answer received: */ 
							if (SDLNet_TCP_Recv(gameinfo->server_socket,data,4)==4 &&
								SDLNet_Read16(data)==18) {
								current_cycle=SDLNet_Read16(data+2);
								last_synchronized_cycle=(current_cycle-1)&0xff;
								connection_state=2;
							} /* if */ 
						} /* if */ 
					} /* if */ 
				} /* if */ 
				break;

		default:
				while(retval && SDLNet_CheckSockets(gameinfo->socket_set, LISTENING_TIME)!=0) {
					new_lvk=false;

					if (SDLNet_SocketReady(gameinfo->server_socket)) {
						if (SDLNet_TCP_Recv(gameinfo->server_socket,data,4)!=4) {
							retval=false;
						} else {
							switch(SDLNet_Read16(data)) {
							case 26:
								// Server Disconnectiong:
								retval=false;
							break;

							case 20:
								{
									int i,pos;
									int cycle,id;
									KEYBOARDSTATE *lvk;

									new_lvk=true;

									/* Receiving an LVK from the server: */ 
									msg_len=SDLNet_Read16(data+2);
									if (SDLNet_TCP_Recv(gameinfo->server_socket,data,msg_len)!=msg_len) {
										retval=false;
									} else {
										id=SDLNet_Read16(data);
										cycle=SDLNet_Read16(data+2);
										pos=4;
										lvk=LVK[id];

										for(i=0;i<SDLK_LAST;i++) {
											lvk->old_keyboard[i]=lvk->keyboard[i];
										} /* for */ 
										LVK_cycle[id]=cycle;
										lvk->keyevents.Delete();
										
										while(pos<msg_len) {
											switch(data[pos++]) {
											case 0:i=SDLNet_Read16(data+pos);
												   pos+=2;
												   lvk->keyboard[i]=1;
												break;
											case 1:i=SDLNet_Read16(data+pos);
												   pos+=2;
												   lvk->keyboard[i]=0;
												break;
											case 2:{
														SDL_keysym *ks;

														ks=new SDL_keysym();
														ks->scancode=(unsigned char)SDLNet_Read16(data+pos);
														ks->sym=(SDLKey)SDLNet_Read16(data+pos+2);
														ks->mod=(SDLMod)SDLNet_Read16(data+pos+4);
														ks->unicode=SDLNet_Read16(data+pos+6);
														lvk->keyevents.Add(ks);
														pos+=8;
												   }
												break;
											} /* switch */ 
										} /* while */ 	
									} /* if */ 
								}
								break;
							} /* switch */ 
						} /* if */ 
					} /* if */ 

					// Update wrong hypotheses: 
					if (new_lvk) {
						bool all_present=true;
						int i;

						for(i=0;i<gameinfo->clients.Length()+1;i++) {
							if (LVK_cycle[i]==last_synchronized_cycle) all_present=false;
						} /* for */ 

						if (all_present) {
							int next_cycle=(last_synchronized_cycle+1)&0xff;
							KEYBOARDSTATE **LVKs;

							LVKs=new KEYBOARDSTATE *[gameinfo->clients.Length()+1];
							for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]=LVK[i];

							LVKs[gameinfo->local->LVK_ID]=hypothetic_LVK[gameinfo->local->LVK_ID][0];

							game_update_hypotheses(LVKs,gameinfo->clients.Length()+1,next_cycle);

							for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]=0;

							if (game_mode==2) {
								fprintf(replay_file,"NEWCYCLE\n",next_cycle);
								for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]->save(replay_file);
							} /* if */ 

							delete LVKs;

							delete hypothetic_LVK[gameinfo->local->LVK_ID].ExtractIni();
							last_synchronized_cycle=next_cycle;
						} /* if */ 
					} /* if */ 

				} /* while */ 

		} /* switch */ 

	} /* if */ 

	return retval;
} /* NetworkMOG2Protocol::listening_cycle */ 



bool NetworkMOG2Protocol::execute_cycle(void)
{
	bool retval=true;
	char data[MSG_MAXLEN];
	KEYBOARDSTATE *local_LVK=LVK[gameinfo->local->LVK_ID];

	// Execute an hypotetical cycle:
	if (gameinfo->server==gameinfo->local) {
		/* SERVER CYCLE: */ 
		if (!bloqued &&
			((current_cycle-last_synchronized_cycle)&0xff)<MAX_HYPOTHESES) {

			// Execute an LVK cycle: 
			{
				int i;
				unsigned char *keyboard = (unsigned char *)SDL_GetKeyState(NULL);
				for(i=0;i<SDLK_LAST;i++) {
					local_LVK->old_keyboard[i]=local_LVK->keyboard[i];
					local_LVK->keyboard[i]=keyboard[i];
				} /* for */ 
				LVK_cycle[local_LVK->LVK_ID]=current_cycle;
			}

			// Execute an hypothetical cycle: 
			/* The SERVER has to execute an hypothetical cycle	*/ 
			/* using the LVKs contained in the "LVK" array, but	*/ 
			/* if some of these LVKs have a higher cycle identi-*/ 
			/* fier than the current cycle, then an older LVK	*/ 
			/* from the "others_LVK" has to be used.			*/ 
			{
				int i;
				KEYBOARDSTATE **LVKs;

				LVKs=new KEYBOARDSTATE *[gameinfo->clients.Length()+1];

				for(i=0;i<gameinfo->clients.Length()+1;i++) {
					if (LVK_cycle[i]!=current_cycle &&
						((LVK_cycle[i]-current_cycle)&0xff)<MAX_HYPOTHESES) LVKs[i]=hypothetic_LVK[i][0];
																	   else LVKs[i]=LVK[i];
				} /* for */ 

				game_hypothetic_cycle(LVKs,gameinfo->clients.Length()+1,current_cycle);

				for(i=0;i<gameinfo->clients.Length()+1;i++) LVKs[i]=0;
				delete LVKs;
			}

			{
				KEYBOARDSTATE *lvk;

				lvk=new KEYBOARDSTATE();
				lvk->copy(local_LVK);
				hypothetic_LVK[local_LVK->LVK_ID].Add(lvk);
				local_LVK->keyevents.Delete();
			}
			current_cycle=(current_cycle+1)&0xff;
		} else {

			/* If the computer is bloqued, wait until some cycles are received, and then continue: */ 
			if (bloqued) {
				if (((current_cycle-last_synchronized_cycle)&0xff)<MAX_HYPOTHESES/2) bloqued=false;
			} else {
				bloqued=true;
			} /* if */ 
		} /* if */ 

	} else {
		switch(connection_state) {
		case 0:
			// Client still not registered: 
			break;

		case 1:
			// Awaiting Server confirmation: 
			break;

		default:
			/* CLIENT CYCLE: */ 
			if (!bloqued &&
				((current_cycle-last_synchronized_cycle)&0xff)<MAX_HYPOTHESES) {

				// Execute an LVK cycle: 
				{
					int i;
					unsigned char *keyboard = (unsigned char *)SDL_GetKeyState(NULL);
					KEYBOARDSTATE *lvk;
					for(i=0;i<SDLK_LAST;i++) {
						local_LVK->old_keyboard[i]=local_LVK->keyboard[i];
						local_LVK->keyboard[i]=keyboard[i];
					} /* for */ 
					LVK_cycle[local_LVK->LVK_ID]=current_cycle;

					lvk=new KEYBOARDSTATE();
					lvk->copy(local_LVK);
					hypothetic_LVK[local_LVK->LVK_ID].Add(lvk);
				}

				/* Execute hypothetic cycle: */ 
				game_hypothetic_cycle(LVK,gameinfo->clients.Length()+1,current_cycle);

				// Send client LVK:
				{
					int i;
					int pos;
					List<SDL_keysym> l;
					SDL_keysym *ks;

					SDLNet_Write16(20,data);
					SDLNet_Write16(local_LVK->LVK_ID,data+4);
					SDLNet_Write16(current_cycle,data+6);
					pos=8;

					for(i=0;i<SDLK_LAST;i++) {
						if (local_LVK->keyboard[i]!=local_LVK->old_keyboard[i]) {
							if (local_LVK->keyboard[i] && !local_LVK->old_keyboard[i]) {
								data[pos++]=0;
								SDLNet_Write16(i,data+pos);
								pos+=2;
							} else {
								data[pos++]=1;
								SDLNet_Write16(i,data+pos);
								pos+=2;
							} /* if */ 
						} /* if */ 
					} /* for */ 

					l.Instance(local_LVK->keyevents);
					l.Rewind();
					while(l.Iterate(ks)) {
						data[pos++]=2;
						SDLNet_Write16(ks->scancode,data+pos);
						SDLNet_Write16(ks->sym,data+pos+2);
						SDLNet_Write16(ks->mod,data+pos+4);
						SDLNet_Write16(ks->unicode,data+pos+6);
						pos+=8;
					} /* while */ 

					SDLNet_Write16(pos-4,data+2);
					SDLNet_TCP_Send(gameinfo->server_socket,data,pos);
				}
				local_LVK->keyevents.Delete();
				current_cycle=(current_cycle+1)&0xff;

			} else {

				/* If the computer is bloqued, wait until some cycles are received, and then continue: */ 
				if (bloqued) {
					if (((current_cycle-last_synchronized_cycle)&0xff)<MAX_HYPOTHESES/2) {
						bloqued=false;
					} /* if */ 
				} else {
					bloqued=true;
				} /* if */ 
			} /* if */ 
						
		} /* switch */ 

		/* End of LVK cycle: */ 
		local_LVK->keyevents.Delete();
	} /* if */ 

	return retval;
} /* NetworkMOG2Protocol::execute_cycle */ 


bool NetworkMOG2Protocol::event(SDL_Event *event)
{
	if (event->type==SDL_KEYDOWN) {
		/* Keyboard event */ 
		SDL_keysym *ks;
		ks=new SDL_keysym();
		*ks=event->key.keysym;
		LVK[gameinfo->local->LVK_ID]->keyevents.Add(ks);
    } /* if */ 

	return true;
} /* NetworkMOG2Protocol::Event */ 


bool NetworkMOG2Protocol::game_hypothetic_cycle(KEYBOARDSTATE **,int ,int )
{
	return true;
} /* NetworkMOG2Protocol::game_hypothetic_cycle */ 


bool NetworkMOG2Protocol::game_update_hypotheses(KEYBOARDSTATE **,int ,int )
{
	return true;
} /* NetworkMOG2Protocol::game_update_hypotheses */ 



ActiveNetworkMOG2::ActiveNetworkMOG2(NWMOG2GameInfo *gi,int gm,char *file) : NetworkMOG2Protocol(gi,gm,file)
{
	SDL_Rect vp;
	ActiveOBJECT *o;
	NWMOG2PlayerInfo *pi;
	List<NWMOG2PlayerInfo> l;
	NWMOG2ClientInfo *ci;
	List<NWMOG2ClientInfo> l2;

	game=new ActiveMOG2(gi->game);

	l2.Instance(gi->clients);
	l2.Rewind();
	while(l2.Iterate(ci)) {
		l.Instance(ci->players);
		l.Rewind();
		while(l.Iterate(pi)) {
			vp=pi->vp_area;
			o=game->add_player(pi->start_interface,pi->start_room,pi->avatar->name,pi->name,&vp,pi->vp_zoom,ci->LVK_ID);
			o->left_key=pi->left_key;
			o->right_key=pi->right_key;
			o->up_key=pi->up_key;
			o->down_key=pi->down_key;
			o->fire1_key=pi->fire1_key;
			o->fire2_key=pi->fire2_key;
			o->fire3_key=pi->fire3_key;
			o->fire4_key=pi->fire4_key;
		} /* while */ 
	} /* while */ 

	ci=gi->server;
	l.Instance(ci->players);
	l.Rewind();
	while(l.Iterate(pi)) {
		vp=pi->vp_area;
		o=game->add_player(pi->start_interface,pi->start_room,pi->avatar->name,pi->name,&vp,pi->vp_zoom,ci->LVK_ID);
		o->left_key=pi->left_key;
		o->right_key=pi->right_key;
		o->up_key=pi->up_key;
		o->down_key=pi->down_key;
		o->fire1_key=pi->fire1_key;
		o->fire2_key=pi->fire2_key;
		o->fire3_key=pi->fire3_key;
		o->fire4_key=pi->fire4_key;
	} /* while */ 


} /* ActiveNetworkMOG2::ActiveNetworkMOG2 */ 


bool ActiveNetworkMOG2::game_hypothetic_cycle(KEYBOARDSTATE **LVK,int n_LVK,int n_cycle)
{

	return true;
} /* ActiveNetworkMOG2::game_hypothetic_cycle */ 


bool ActiveNetworkMOG2::game_update_hypotheses(KEYBOARDSTATE **LVK,int n_LVK,int n_cycle)
{
	game->cycle(LVK,n_LVK);
	return true;
} /* ActiveNetworkMOG2::game_update_hypotheses */ 
	
